@@ -1,8 +1,9 @@
 import argparse
 import asyncio
+import pathlib
+import sqlite3
 from pathlib import Path
 
-import yarl
 
 from . import __version__ as VERSION
 from .utils.scraper import scrape
@@ -18,25 +19,19 @@ def parse_args():
     parser.add_argument("--log-file", help="log file to write to", default="downloader.log")
     parser.add_argument("--threads", type=int, help="number of threads to use (0 = max)", default=0)
     parser.add_argument("--attempts", type=int, help="number of attempts to download each file", default=10)
+    parser.add_argument("--disable-attempt-limit", help="disables the attempt limitation", action="store_true")
     parser.add_argument("--include-id", help="include the ID in the download folder name", action="store_true")
     parser.add_argument("--exclude-videos", help="skip downloading of video files", action="store_true")
     parser.add_argument("--exclude-images", help="skip downloading of image files", action="store_true")
     parser.add_argument("--exclude-audio", help="skip downloading of audio files", action="store_true")
     parser.add_argument("--exclude-other", help="skip downloading of images", action="store_true")
-    parser.add_argument("links", metavar="link", nargs="*",
-                        help="link to content to download (passing multiple links is supported)", default=[])
+    parser.add_argument("--thotsbay-username", type=str, help="username to login to thotsbay", default=None)
+    parser.add_argument("--thotsbay-password", type=str, help="password to login to thotsbay", default=None)
+    parser.add_argument("links", metavar="link", nargs="*", help="link to content to download (passing multiple links is supported)", default=[])
     args = parser.parse_args()
     return args
 
 
-def regex_links(urls) -> list:
-    all_links = [x.group().replace(".md.", ".") for x in re.finditer(r"(?:http.*?)(?=('|$|\n|\r\n|\r|\s|\"|\[/URL]))", urls)]
-    yarl_links = []
-    for link in all_links:
-        yarl_links.append(yarl.URL(link))
-    return yarl_links
-
-
 async def download_all(args: argparse.Namespace):
     clear()
     log(f"We are running version {VERSION} of Cyberdrop Downloader", Fore.WHITE)
@@ -47,10 +42,12 @@ async def download_all(args: argparse.Namespace):
         log(f"{input_file} created. Populate it and retry.")
         exit(1)
 
+    conn, curr = sql_initialize()
+
     links = args.links
     with open(input_file, "r") as f:
         links += regex_links(f.read())
-    content_object = await scrape(links, args.include_id)
+    content_object = await scrape(links, args.include_id, args.thotsbay_username, args.thotsbay_password)
     if content_object.is_empty():
         logging.error(f'ValueError No links')
         log('No links found, check the URL.txt\nIf the link works in your web browser, '
@@ -59,16 +56,21 @@ async def download_all(args: argparse.Namespace):
         exit(0)
     clear()
     downloaders = get_downloaders(content_object, folder=Path(args.output_folder), attempts=args.attempts,
+                                  disable_attempt_limit=args.disable_attempt_limit,
                                   threads=args.threads, exclude_videos=args.exclude_videos,
                                   exclude_images=args.exclude_images, exclude_audio=args.exclude_audio,
-                                  exclude_other=args.exclude_other)
+                                  exclude_other=args.exclude_other, connection=conn, cursor=curr)
 
     for downloader in downloaders:
         await downloader.download_content()
     logger.debug("Finished")
+
+    all_files = [str(f) for f in pathlib.Path(args.output_folder).glob("**/*") if f.is_file()]
+    combined = '\t'.join(all_files)
+
     log('Finished downloading. Enjoy :)')
-    log('If you have ".download" files remaining, rerun this program. '
-        'You most likely ran into download attempts limits')
+    if '.part' in combined:
+        log('There are still partial downloads in your folders, please re-run the program.')
 
 
 def main():
