@@ -1,17 +1,13 @@
 import asyncio
 from functools import wraps
-import http
 import multiprocessing
 from pathlib import Path
-import ssl
 import time
 import traceback
 
 import aiofiles
 import aiofiles.os
-import aiohttp
 import aiohttp.client_exceptions
-import certifi
 from tqdm import tqdm
 
 from .base_functions import *
@@ -26,14 +22,16 @@ class FailureException(Exception):
 def retry(f):
     @wraps(f)
     async def wrapper(self, *args, **kwargs):
-        for i in range(self.attempts):
+        while True:
             try:
                 return await f(self, *args, **kwargs)
             except FailureException:
-                if i >= self.attempts - 1:
-                    raise
+                if not self.disable_attempt_limit:
+                    if self.current_attempt >= self.attempts - 1:
+                        raise
                 logger.debug('Retrying...')
-                time.sleep(2)
+                self.attempts += 1
+                await asyncio.sleep(2)
     return wrapper
 
 
@@ -65,13 +63,16 @@ async def throttle(self, url: URL) -> None:
 
 
 class Downloader:
-    def __init__(self, album_obj: AlbumItem, morsels, folder: Path, title: str, attempts: int, max_workers: int,
-                 exclude_videos: bool, exclude_images: bool, exclude_audio: bool, exclude_other: bool):
+    def __init__(self, album_obj: AlbumItem, cookie_jar, folder: Path, title: str, attempts: int,
+                 disable_attempt_limit: bool, max_workers: int, exclude_videos: bool, exclude_images: bool,
+                 exclude_audio: bool, exclude_other: bool, connection: sqlite3.Connection, cursor: sqlite3.Cursor):
         self.album_obj = album_obj
-        self.morsels = morsels
+        self.cookie_jar = cookie_jar
         self.folder = folder
         self.title = title
         self.attempts = attempts
+        self.current_attempt = 0
+        self.disable_attempt_limit = disable_attempt_limit
         self.max_workers = max_workers
         self.exclude_videos = exclude_videos
         self.exclude_images = exclude_images
@@ -80,6 +81,8 @@ class Downloader:
         self._semaphore = asyncio.Semaphore(max_workers)
         self.delay = {'media-files.bunkr.is': 2}
         self.throttle_times = {}
+        self.connection = connection
+        self.cursor = cursor
 
     """Changed from aiohttp exceptions caught to FailureException to allow for partial downloads."""
 
@@ -94,21 +97,20 @@ class Downloader:
     ) -> None:
         """Download the content of given URL"""
         headers = {'Referer': str(referral), 'user-agent': user_agent}
-        ssl_context = ssl.create_default_context(cafile=certifi.where())
 
         try:
             async with self._semaphore:
                 if url.host in self.delay:
                     await throttle(self, url)
 
-                ext = '.'+filename.split('.')[-1]
+                ext = '.'+filename.split('.')[-1].lower()
                 if not (ext in FILE_FORMATS['Images'] or ext in FILE_FORMATS['Videos'] or ext in FILE_FORMATS['Audio'] or ext in FILE_FORMATS['Other']):
-                    resp = await session.get(url, headers=headers, ssl=ssl_context, raise_for_status=True)
-                    filename = resp.content_disposition.filename
-                    filename = sanitize(filename)
-                    del resp
-                    if (self.folder / self.title / filename).exists():
-                        return
+                    async with session.get(url, headers=headers, ssl=ssl_context, raise_for_status=True) as resp:
+                        filename = resp.content_disposition.filename
+                        filename = sanitize(filename)
+                        del resp
+                        if (self.folder / self.title / filename).exists():
+                            return
 
                 # Skip based on CLI arg.
                 ext = '.' + filename.split('.')[-1]
@@ -129,6 +131,13 @@ class Downloader:
                         logging.debug("Skipping " + filename)
                         return
 
+                if await sql_check_existing(self.cursor, str(url)):
+                    log("\n%s Already Downloaded" % filename)
+                    logger.debug("%s was found in db file (Previously downloaded)" % filename)
+                    return
+
+                await sql_insert_file(self.connection, self.cursor, url, filename, 0)
+
                 resume_point = 0
                 complete_file = (self.folder / self.title / filename)
                 temp_file = complete_file.with_suffix(complete_file.suffix + '.part')
@@ -137,26 +146,31 @@ class Downloader:
                     resume_point = temp_file.stat().st_size
                     headers['Range'] = 'bytes=%d-' % resume_point
 
-                resp = await session.get(url, headers=headers, ssl=ssl_context, raise_for_status=True)
-
-                content_type = resp.headers.get('Content-Type')
-                if 'text' in content_type.lower() or 'html' in content_type.lower():
-                    log(f"\nServer for {url} is either down or the file no longer exists", Fore.RED)
-                    return
-
-                total = int(resp.headers.get('Content-Length', str(0))) + resume_point
-                with tqdm(
-                        total=total, unit_scale=True,
-                        unit='B', leave=False, initial=resume_point,
-                        desc=filename, disable=(not show_progress)
-                ) as progress:
-                    async with aiofiles.open(temp_file, mode='ab') as f:
-                        async for chunk, _ in resp.content.iter_chunks():
-                            await f.write(chunk)
-                            progress.update(len(chunk))
+                async with session.get(url, headers=headers, ssl=ssl_context, raise_for_status=True, allow_redirects=False) as resp:
+                    content_type = resp.headers.get('Content-Type')
+                    if 'text' in content_type.lower() or 'html' in content_type.lower():
+                        log(f"\nServer for {url} is either down or the file no longer exists", Fore.RED)
+                        return
+                    total = int(resp.headers.get('Content-Length', str(0))) + resume_point
+                    with tqdm(
+                            total=total, unit_scale=True,
+                            unit='B', leave=False, initial=resume_point,
+                            desc=filename, disable=(not show_progress)
+                    ) as progress:
+                        async with aiofiles.open(temp_file, mode='ab') as f:
+                            async for chunk, _ in resp.content.iter_chunks():
+                                await f.write(chunk)
+                                progress.update(len(chunk))
+            await sql_update_file(self.connection, self.cursor, url, filename, 1)
+            resp.close()
             await self.rename_file(filename)
         except (aiohttp.client_exceptions.ClientPayloadError, aiohttp.client_exceptions.ClientOSError,
-                aiohttp.client_exceptions.ServerDisconnectedError, asyncio.TimeoutError, FailureException) as e:
+                aiohttp.client_exceptions.ServerDisconnectedError, asyncio.TimeoutError,
+                aiohttp.client_exceptions.ClientResponseError, FailureException) as e:
+            try:
+                resp.close()
+            except:
+                pass
             raise FailureException(e)
 
     async def rename_file(self, filename: str) -> None:
@@ -189,6 +203,7 @@ class Downloader:
 
         if (self.folder / self.title / filename).exists():
             logger.debug(str(self.folder / self.title / filename) + " Already Exists")
+            await sql_update_file(self.connection, self.cursor, url, filename, 1)
         else:
             logger.debug("Working on " + str(url))
             try:
@@ -196,8 +211,7 @@ class Downloader:
                                          session=session, show_progress=show_progress)
             except Exception:
                 logger.debug(traceback.format_exc())
-                log(f"\nSkipping {filename}: likely exceeded download attempts (or ran into an error)\nRe-run program "
-                    f"after exit to continue download.\n", Fore.WHITE)
+                log(f"\nError attempting {filename}: See downloader.log for details\n", Fore.WHITE)
 
     async def download_all(
             self,
@@ -217,32 +231,12 @@ class Downloader:
     ) -> None:
         """Download the content of all links and save them as files."""
         (self.folder / self.title).mkdir(parents=True, exist_ok=True)
-        async with aiohttp.ClientSession() as session:
-            session.cookie_jar.update_cookies(self.morsels)
+        async with aiohttp.ClientSession(cookie_jar=self.cookie_jar) as session:
             await self.download_all(self.album_obj, session, show_progress=show_progress)
 
 
-def simple_cookies(cookies):
-    morsels = {}
-    for cookie in cookies:
-        # https://docs.python.org/3/library/http.cookies.html#morsel-objects
-        morsel = http.cookies.Morsel()
-        morsel.set(cookie["name"], cookie["value"], cookie["value"])
-        if "domain" in cookie:
-            morsel["domain"] = cookie["domain"]
-        if "httpOnly" in cookie:
-            morsel["httponly"] = cookie["httpOnly"]
-        if "path" in cookie:
-            morsel["path"] = cookie["path"]
-        if "secure" in cookie:
-            morsel["secure"] = cookie["secure"]
-
-        morsels[cookie["name"]] = morsel
-    return morsels
-
-
-def get_downloaders(Cascade: CascadeItem, folder: Path, attempts: int, threads: int, exclude_videos: bool,
-                    exclude_images: bool, exclude_audio: bool, exclude_other: bool) -> List[Downloader]:
+def get_downloaders(Cascade: CascadeItem, folder: Path, attempts: int, disable_attempt_limit: bool, threads: int, exclude_videos: bool,
+                    exclude_images: bool, exclude_audio: bool, exclude_other: bool, connection:sqlite3.Connection, cursor: sqlite3.Cursor) -> List[Downloader]:
     """Get a list of downloaders for each supported type of URLs.
     We shouldn't just assume that each URL will have the same netloc as
     the first one, so we need to classify them one by one, sort them to
@@ -251,16 +245,18 @@ def get_downloaders(Cascade: CascadeItem, folder: Path, attempts: int, threads:
     """
 
     downloaders = []
-    morsels = simple_cookies(Cascade.cookies.cookies)
+
+    cookie_jar = Cascade.cookies
 
     for domain, domain_obj in Cascade.domains.items():
         max_workers = threads if threads != 0 else multiprocessing.cpu_count()
         if 'bunkr' in domain:
             max_workers = 2 if (max_workers > 2) else max_workers
         for title, album_obj in domain_obj.albums.items():
-            downloader = Downloader(album_obj, morsels=morsels, title=title, folder=folder,
-                                    attempts=attempts, max_workers=max_workers, exclude_videos=exclude_videos,
+            downloader = Downloader(album_obj, cookie_jar=cookie_jar, title=title, folder=folder,
+                                    attempts=attempts, disable_attempt_limit=disable_attempt_limit,
+                                    max_workers=max_workers, exclude_videos=exclude_videos,
                                     exclude_images=exclude_images, exclude_audio=exclude_audio,
-                                    exclude_other=exclude_other)
+                                    exclude_other=exclude_other, connection=connection, cursor=cursor)
             downloaders.append(downloader)
     return downloaders
